#include "src/storage/compression/numeric_constant.cpp"

#include "src/storage/compression/fixed_size_uncompressed.cpp"

#include "src/storage/compression/rle.cpp"

#include "src/storage/compression/dictionary_compression.cpp"

#include "src/storage/compression/string_uncompressed.cpp"

#include "src/storage/compression/uncompressed.cpp"

#include "src/storage/compression/validity_uncompressed.cpp"

#include "src/storage/compression/bitpacking.cpp"

#include "src/storage/compression/bitpacking_hugeint.cpp"

#include "src/storage/compression/patas.cpp"

#include "src/storage/compression/alprd.cpp"

#include "src/storage/compression/fsst.cpp"

