#include "src/optimizer/statistics/expression/propagate_aggregate.cpp"

#include "src/optimizer/statistics/expression/propagate_between.cpp"

#include "src/optimizer/statistics/expression/propagate_case.cpp"

#include "src/optimizer/statistics/expression/propagate_cast.cpp"

#include "src/optimizer/statistics/expression/propagate_comparison.cpp"

#include "src/optimizer/statistics/expression/propagate_conjunction.cpp"

#include "src/optimizer/statistics/expression/propagate_constant.cpp"

#include "src/optimizer/statistics/expression/propagate_columnref.cpp"

#include "src/optimizer/statistics/expression/propagate_function.cpp"

#include "src/optimizer/statistics/expression/propagate_operator.cpp"

