#include "src/core_functions/scalar/string/ascii.cpp"

#include "src/core_functions/scalar/string/bar.cpp"

#include "src/core_functions/scalar/string/chr.cpp"

#include "src/core_functions/scalar/string/damerau_levenshtein.cpp"

#include "src/core_functions/scalar/string/format_bytes.cpp"

#include "src/core_functions/scalar/string/hamming.cpp"

#include "src/core_functions/scalar/string/hex.cpp"

#include "src/core_functions/scalar/string/instr.cpp"

#include "src/core_functions/scalar/string/jaccard.cpp"

#include "src/core_functions/scalar/string/jaro_winkler.cpp"

#include "src/core_functions/scalar/string/left_right.cpp"

#include "src/core_functions/scalar/string/levenshtein.cpp"

#include "src/core_functions/scalar/string/md5.cpp"

#include "src/core_functions/scalar/string/pad.cpp"

#include "src/core_functions/scalar/string/parse_path.cpp"

#include "src/core_functions/scalar/string/printf.cpp"

#include "src/core_functions/scalar/string/regexp_escape.cpp"

#include "src/core_functions/scalar/string/repeat.cpp"

#include "src/core_functions/scalar/string/replace.cpp"

#include "src/core_functions/scalar/string/reverse.cpp"

#include "src/core_functions/scalar/string/sha256.cpp"

#include "src/core_functions/scalar/string/starts_with.cpp"

#include "src/core_functions/scalar/string/string_split.cpp"

#include "src/core_functions/scalar/string/to_base.cpp"

#include "src/core_functions/scalar/string/translate.cpp"

#include "src/core_functions/scalar/string/trim.cpp"

#include "src/core_functions/scalar/string/unicode.cpp"

