#include "src/parser/tableref/basetableref.cpp"

#include "src/parser/tableref/emptytableref.cpp"

#include "src/parser/tableref/expressionlistref.cpp"

#include "src/parser/tableref/joinref.cpp"

#include "src/parser/tableref/pivotref.cpp"

#include "src/parser/tableref/showref.cpp"

#include "src/parser/tableref/subqueryref.cpp"

#include "src/parser/tableref/table_function.cpp"

