#include "src/execution/physical_plan/plan_aggregate.cpp"

#include "src/execution/physical_plan/plan_any_join.cpp"

#include "src/execution/physical_plan/plan_asof_join.cpp"

#include "src/execution/physical_plan/plan_column_data_get.cpp"

#include "src/execution/physical_plan/plan_comparison_join.cpp"

#include "src/execution/physical_plan/plan_copy_database.cpp"

#include "src/execution/physical_plan/plan_copy_to_file.cpp"

#include "src/execution/physical_plan/plan_create.cpp"

#include "src/execution/physical_plan/plan_create_index.cpp"

#include "src/execution/physical_plan/plan_create_table.cpp"

#include "src/execution/physical_plan/plan_create_secret.cpp"

#include "src/execution/physical_plan/plan_cross_product.cpp"

#include "src/execution/physical_plan/plan_delete.cpp"

#include "src/execution/physical_plan/plan_delim_get.cpp"

#include "src/execution/physical_plan/plan_delim_join.cpp"

#include "src/execution/physical_plan/plan_distinct.cpp"

#include "src/execution/physical_plan/plan_dummy_scan.cpp"

#include "src/execution/physical_plan/plan_empty_result.cpp"

#include "src/execution/physical_plan/plan_execute.cpp"

#include "src/execution/physical_plan/plan_explain.cpp"

#include "src/execution/physical_plan/plan_export.cpp"

#include "src/execution/physical_plan/plan_filter.cpp"

#include "src/execution/physical_plan/plan_get.cpp"

#include "src/execution/physical_plan/plan_insert.cpp"

#include "src/execution/physical_plan/plan_limit.cpp"

#include "src/execution/physical_plan/plan_limit_percent.cpp"

#include "src/execution/physical_plan/plan_order.cpp"

#include "src/execution/physical_plan/plan_pivot.cpp"

#include "src/execution/physical_plan/plan_positional_join.cpp"

#include "src/execution/physical_plan/plan_pragma.cpp"

#include "src/execution/physical_plan/plan_prepare.cpp"

#include "src/execution/physical_plan/plan_projection.cpp"

#include "src/execution/physical_plan/plan_set_operation.cpp"

#include "src/execution/physical_plan/plan_set.cpp"

#include "src/execution/physical_plan/plan_reset.cpp"

#include "src/execution/physical_plan/plan_sample.cpp"

#include "src/execution/physical_plan/plan_simple.cpp"

#include "src/execution/physical_plan/plan_top_n.cpp"

#include "src/execution/physical_plan/plan_update.cpp"

#include "src/execution/physical_plan/plan_window.cpp"

#include "src/execution/physical_plan/plan_unnest.cpp"

#include "src/execution/physical_plan/plan_expression_get.cpp"

#include "src/execution/physical_plan/plan_recursive_cte.cpp"

#include "src/execution/physical_plan/plan_cte.cpp"

