/*
THIS IS AN AUTOMATICALLY GENERATED FILE
DO NOT EDIT

See distcomp.c for details
*/

/* aliases for values/weights in the first_names distribution */
#define first_names_name    1
#define first_names_male    1
#define first_names_female  2
#define first_names_unified 3

/* aliases for values/weights in the last_names distribution */
/* NONE DEFINED */

/* aliases for values/weights in the calendar distribution */
#define calendar_day_seq      1
#define calendar_month_name   2
#define calendar_date         3
#define calendar_season       4
#define calendar_month_num    5
#define calendar_quarter      6
#define calendar_fom          7
#define calendar_holiday      8
#define calendar_uniform      1
#define calendar_uniform_leap 2
#define calendar_sales        3
#define calendar_sales_leap   4
#define calendar_returns      5
#define calendar_returns_leap 6
#define calendar_skewed       7
#define calendar_low          8
#define calendar_medium       9
#define calendar_high         10

/* aliases for values/weights in the week_info distribution */
/* NONE DEFINED */

/* aliases for values/weights in the fips_county distribution */
#define fips_county_fips       1
#define fips_county_county     2
#define fips_county_st         3
#define fips_county_state      4
#define fips_county_zone       5
#define fips_county_gmt        6
#define fips_county_uniform    1
#define fips_county_population 2
#define fips_county_tz         3
#define fips_county_tz90       4
#define fips_county_tz9        5
#define fips_county_tz1        6

/* aliases for values/weights in the street_names distribution */
/* NONE DEFINED */

/* aliases for values/weights in the street_type distribution */
/* NONE DEFINED */

/* aliases for values/weights in the adjectives distribution */
/* NONE DEFINED */

/* aliases for values/weights in the adverbs distribution */
/* NONE DEFINED */

/* aliases for values/weights in the articles distribution */
/* NONE DEFINED */

/* aliases for values/weights in the nouns distribution */
/* NONE DEFINED */

/* aliases for values/weights in the prepositions distribution */
/* NONE DEFINED */

/* aliases for values/weights in the verbs distribution */
/* NONE DEFINED */

/* aliases for values/weights in the auxiliaries distribution */
/* NONE DEFINED */

/* aliases for values/weights in the terminators distribution */
/* NONE DEFINED */

/* aliases for values/weights in the sentences distribution */
/* NONE DEFINED */

/* aliases for values/weights in the syllables distribution */
/* NONE DEFINED */

/* aliases for values/weights in the cities distribution */
#define cities_name    1
#define cities_usgs    1
#define cities_uniform 2
#define cities_large   3
#define cities_medium  4
#define cities_small   5
#define cities_unified 6

/* aliases for values/weights in the categories distribution */
/* NONE DEFINED */

/* aliases for values/weights in the women_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the men_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the children_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the shoe_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the music_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the jewelry_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the home_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the sport_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the book_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the electronic_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the sizes distribution */
/* NONE DEFINED */

/* aliases for values/weights in the units distribution */
/* NONE DEFINED */

/* aliases for values/weights in the container distribution */
/* NONE DEFINED */

/* aliases for values/weights in the colors distribution */
#define colors_name    1
#define colors_uniform 1
#define colors_skewed  2
#define colors_low     3
#define colors_medium  4
#define colors_high    5

/* aliases for values/weights in the brand_syllables distribution */
/* NONE DEFINED */

/* aliases for values/weights in the i_current_price distribution */
#define i_current_price_index      1
#define i_current_price_low_bound  2
#define i_current_price_high_bound 3
#define i_current_price_skew       1
#define i_current_price_high       2
#define i_current_price_medium     3
#define i_current_price_low        4

/* aliases for values/weights in the i_manufact_id distribution */
/* NONE DEFINED */

/* aliases for values/weights in the i_manager_id distribution */
/* NONE DEFINED */

/* aliases for values/weights in the rowcounts distribution */
/* NONE DEFINED */

/* aliases for values/weights in the gender distribution */
/* NONE DEFINED */

/* aliases for values/weights in the marital_status distribution */
/* NONE DEFINED */

/* aliases for values/weights in the education distribution */
/* NONE DEFINED */

/* aliases for values/weights in the purchase_band distribution */
/* NONE DEFINED */

/* aliases for values/weights in the buy_potential distribution */
/* NONE DEFINED */

/* aliases for values/weights in the credit_rating distribution */
/* NONE DEFINED */

/* aliases for values/weights in the income_band distribution */
/* NONE DEFINED */

/* aliases for values/weights in the dependent_count distribution */
/* NONE DEFINED */

/* aliases for values/weights in the vehicle_count distribution */
/* NONE DEFINED */

/* aliases for values/weights in the promo_purpose distribution */
/* NONE DEFINED */

/* aliases for values/weights in the return_reasons distribution */
/* NONE DEFINED */

/* aliases for values/weights in the store_type distribution */
/* NONE DEFINED */

/* aliases for values/weights in the geography_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the divisions distribution */
/* NONE DEFINED */

/* aliases for values/weights in the stores distribution */
/* NONE DEFINED */

/* aliases for values/weights in the hours distribution */
/* NONE DEFINED */

/* aliases for values/weights in the call_centers distribution */
/* NONE DEFINED */

/* aliases for values/weights in the call_center_hours distribution */
/* NONE DEFINED */

/* aliases for values/weights in the call_center_class distribution */
/* NONE DEFINED */

/* aliases for values/weights in the salutations distribution */
/* NONE DEFINED */

/* aliases for values/weights in the ship_mode_code distribution */
/* NONE DEFINED */

/* aliases for values/weights in the ship_mode_type distribution */
/* NONE DEFINED */

/* aliases for values/weights in the ship_mode_carrier distribution */
/* NONE DEFINED */

/* aliases for values/weights in the web_page_use distribution */
/* NONE DEFINED */

/* aliases for values/weights in the web_page_type distribution */
/* NONE DEFINED */

/* aliases for values/weights in the catalog_page_type distribution */
/* NONE DEFINED */

/* aliases for values/weights in the location_type distribution */
/* NONE DEFINED */

/* aliases for values/weights in the top_domains distribution */
/* NONE DEFINED */

/* aliases for values/weights in the countries distribution */
/* NONE DEFINED */
