#include "src/optimizer/pushdown/pushdown_aggregate.cpp"

#include "src/optimizer/pushdown/pushdown_cross_product.cpp"

#include "src/optimizer/pushdown/pushdown_distinct.cpp"

#include "src/optimizer/pushdown/pushdown_filter.cpp"

#include "src/optimizer/pushdown/pushdown_get.cpp"

#include "src/optimizer/pushdown/pushdown_inner_join.cpp"

#include "src/optimizer/pushdown/pushdown_limit.cpp"

#include "src/optimizer/pushdown/pushdown_left_join.cpp"

#include "src/optimizer/pushdown/pushdown_mark_join.cpp"

#include "src/optimizer/pushdown/pushdown_projection.cpp"

#include "src/optimizer/pushdown/pushdown_semi_anti_join.cpp"

#include "src/optimizer/pushdown/pushdown_set_operation.cpp"

#include "src/optimizer/pushdown/pushdown_single_join.cpp"

