#include "src/storage/arena_allocator.cpp"

#include "src/storage/buffer_manager.cpp"

#include "src/storage/checkpoint_manager.cpp"

#include "src/storage/temporary_memory_manager.cpp"

#include "src/storage/block.cpp"

#include "src/storage/data_pointer.cpp"

#include "src/storage/data_table.cpp"

#include "src/storage/index.cpp"

#include "src/storage/local_storage.cpp"

#include "src/storage/magic_bytes.cpp"

#include "src/storage/storage_manager.cpp"

#include "src/storage/standard_buffer_manager.cpp"

#include "src/storage/table_index_list.cpp"

#include "src/storage/write_ahead_log.cpp"

#include "src/storage/optimistic_data_writer.cpp"

#include "src/storage/partial_block_manager.cpp"

#include "src/storage/single_file_block_manager.cpp"

#include "src/storage/storage_info.cpp"

#include "src/storage/storage_lock.cpp"

#include "src/storage/wal_replay.cpp"

